/* Copyright  1997, 1998, Hagiwara Youichi */
/* All rights reserved.                     */

/* Mrand.c : Maximum-length linearly recurring sequence random number */

/*** #include ***/

#include <stdlib.h>	/* use srand() and rand() */
#include <math.h>	/* use sqrt() */

#include "Mrand.h"	/* include prototype of function implemented in this file */

/*** #define ***/

#define NEW( type )			((type *)malloc( sizeof( type ) ))
#define DELETE( pointer )	(free( (void *)(pointer) ))

/*** typedef ***/

typedef struct MaxSeqRandNum *MaxSeqRandNumPtr;
typedef struct MaxSeqRandNum {
	unsigned long random;
	MaxSeqRandNumPtr next;
} MaxSeqRandNum;

/*** global ***/

static MaxSeqRandNumPtr	seedY23 = NULL, randX23 = NULL, Xt607 = NULL, Xt273 = NULL;
static MaxSeqRandNumPtr	seedY32 = NULL, randX32 = NULL, Xt521 = NULL, Xt32 = NULL;

/* gaussrand */

double gaussrand( double mu, double sigma )
{
	static long double V1, V2, S;
	static short phase = 0;
	long double X;

	if (phase == 0) {
		do {
			long double U1, U2;

			U1 = (long double)rand() / RAND_MAX;
			U2 = (long double)rand() / RAND_MAX;

			V1 = 2 * U1 - 1;
			V2 = 2 * U2 - 1;
			S = V1 * V1 + V2 * V2;
		} while (S >= 1 || S == 0);

		X = V1 * sqrt((double)(-2 * log((double)S) / S));
	} else
		X = V2 * sqrt((double)(-2 * log((double)S) / S));

	phase = 1 - phase;

	return (double)(X * sigma + mu);
}

/*** Msrand23 ***/

int Msrand23( unsigned int seed )
{
	unsigned int	f = 0, s = 0, t = 0;
	unsigned long	seed32to22, seed21to11, seed10to1;

	MaxSeqRandNumPtr	Xfirst = NULL, Xcurrent = NULL, Xlast = NULL,
						Yfirst = NULL, Ycurrent = NULL, Ylast = NULL,
						Yt273 = NULL, Yt607 = NULL,
						Yt19 = NULL, Yt18 = NULL, Yt9 = NULL, Yt8 = NULL;

	srand( seed );

	/* make seedY23 ring */
	/* 0<= t <= 18 */
	for ( t = 0; t < 19; t++ ) {
		if ( seedY23 == NULL ) {
			Ycurrent = NEW( MaxSeqRandNum );
			if ( Ycurrent == NULL ) {
				while ( Yfirst != NULL ) {
					Ycurrent = Yfirst;
					Yfirst = Yfirst->next;
					DELETE( Ycurrent );
				}
				return 1;
			}
			Ycurrent->next = NULL;
			if ( Yfirst == NULL ) {
				Yfirst = Ycurrent;
				Ylast = Ycurrent;
			} else {
				Ylast->next = Ycurrent;
				Ylast = Ycurrent;
			}
		} else if ( Ycurrent != NULL )
			Ycurrent = Ycurrent->next;
		else {
			Ycurrent = seedY23;
			Yfirst = seedY23;
		}
		seed32to22 = (0x7ff0UL & (unsigned long)rand()) >> 17;
		seed21to11 = (0x7ff0UL & (unsigned long)rand()) << 6;
		seed10to1 = (0x7fe0UL & (unsigned long)rand()) >> 5;
		Ycurrent->random = (seed32to22 | seed21to11 | seed10to1);
		switch ( t ) {
			case 0 :
				Yt607 = Ycurrent;
				Yt19 = Ycurrent;
				break;
			case 1 :
				Yt18 = Ycurrent;
				break;
			case 10 :
				Yt9 = Ycurrent;
				break;
			case 11 :
				Yt8 = Ycurrent;
				break;
		}
	}

	/* modify t = 18 */
	Ycurrent->random = (Yfirst->random>>31 ^ Yfirst->random>>17) | (0xffffffffUL & Ycurrent->random<<1);

	/* 19 <= t <= 606 */
	for ( t = 19; t < 607; t++ ) {
		if ( seedY23 == NULL ) {
			Ycurrent = NEW( MaxSeqRandNum );
			if ( Ycurrent == NULL ) {
				while ( Yfirst != NULL ) {
					Ycurrent = Yfirst;
					Yfirst = Yfirst->next;
					DELETE( Ycurrent );
				}
				return 2;
			}
			Ycurrent->next = NULL;
			Ylast->next = Ycurrent;
			Ylast = Ycurrent;
		} else
			Ycurrent = Ycurrent->next;
		Ycurrent->random = 0xffffffffUL & ((Yt19->random<<1 | Yt18->random>>31) ^ (Yt9->random<<15 | Yt8->random>>17));
		if ( t == 334 )
			Yt273 = Ycurrent;
		Yt19 = Yt18;
		Yt18 = Yt18->next;
		Yt9 = Yt8;
		Yt8 = Yt8->next;
	}

	/* ring seedY23 if not ringed */
	if ( seedY23 == NULL ) {
		Ylast->next = Yfirst;
		seedY23 = Yfirst;
	}

	/* set Ycurrent */
	Ycurrent = Yfirst;

	/* make randX23 ring from seedY23 ring */
	for ( f = 0, s = 0, t = 0; t < 6967; f--, t++ ) {
		if ( f == 0 ) {
			if( randX23 == NULL ) {
				Xcurrent = NEW( MaxSeqRandNum );
				if ( Xcurrent == NULL ) {
					for ( t = 0; t < 607; t++ ) {
						Ycurrent = Yfirst;
						Yfirst = Yfirst->next;
						DELETE( Ycurrent );
					}
					while ( Xfirst != NULL ) {
						Xcurrent = Xfirst;
						Xfirst = Xfirst->next;
						DELETE( Xcurrent );
					}
					seedY23 = NULL;
					Xt607 = NULL;
					Xt273 = NULL;
					return 3;
				}
				Xcurrent->next = NULL;
				if ( Xfirst == NULL ) {
					Xfirst = Xcurrent;
					Xlast = Xcurrent;
				} else {
					Xlast->next = Xcurrent;
					Xlast = Xcurrent;
				}
			} else if ( Xcurrent != NULL )
				Xcurrent = Xcurrent->next;
			else {
				Xcurrent = randX23;
				Xfirst = randX23;
			}
			Xcurrent->random = 0x7fffffUL & Ycurrent->random;
			switch ( s ) {
				case 0 :
					Xt607 = Xcurrent;
					break;
				case 334 :
					Xt273 = Xcurrent;
					break;
			}
			s++;
			f = 16;
		}
		Ycurrent->random = Yt273->random ^ Yt607->random;
		Ycurrent = Ycurrent->next;
		Yt273 = Yt273->next;
		Yt607 = Yt607->next;
	}

	/* ring randX23 if not ringed */
	if ( randX23 == NULL ) {
		Xlast->next = Xfirst;
		randX23 = Xfirst;
	}

	return 0;
}

/*** Mrand23 ***/

long Mrand23( void )
{
	long rand;

	if ( randX23 == NULL )
		if ( Msrand23( 1u ) != 0 )	/* if no memory... */
			return -1L;	/* return -1 */

	rand = (signed long)(randX23->random);

	randX23->random = (Xt273->random) ^ (Xt607->random);	/* generate next random */
	randX23 = randX23->next;
	Xt273 = Xt273->next;
	Xt607 = Xt607->next;

	return rand;
}

/* Mgaussrand23 */

double Mgaussrand23( double mu, double sigma )
{
	static long double V1, V2, S;
	static short phase = 0;
	long double X;

	if (phase == 0) {
		do {
			long r1, r2;
			long double U1, U2;
			r1 = Mrand23();
			r2 = Mrand23();
			if (r1 == -1 && r2 == -1)
				return 0;
			U1 = (long double)r1 / MRAND23_MAX;
			U2 = (long double)r2 / MRAND23_MAX;

			V1 = 2 * U1 - 1;
			V2 = 2 * U2 - 1;
			S = V1 * V1 + V2 * V2;
		} while (S >= 1 || S == 0);

		X = V1 * sqrt((double)(-2 * log((double)S) / S));
	} else
		X = V2 * sqrt((double)(-2 * log((double)S) / S));

	phase = 1 - phase;

	return (double)(X * sigma + mu);
}

/*** Msrand32 ***/

int Msrand32( unsigned int seed )
{
	unsigned int	f = 0, s = 0, t = 0;
	unsigned long	seed32to22, seed21to11, seed10to1;

	MaxSeqRandNumPtr	Xfirst = NULL, Xcurrent = NULL, Xlast = NULL,
						Yfirst = NULL, Ycurrent = NULL, Ylast = NULL,
						Yt32 = NULL, Yt521 = NULL,
						Yt17 = NULL, Yt16 = NULL, Yt15 = NULL, Yt1 = NULL;

	srand( seed );

	/* make seedY32 ring */
	/* 0<= t <= 16 */
	for ( t = 0; t < 17; t++ ) {
		if ( seedY32 == NULL ) {
			Ycurrent = NEW( MaxSeqRandNum );
			if ( Ycurrent == NULL ) {
				while ( Yfirst != NULL ) {
					Ycurrent = Yfirst;
					Yfirst = Yfirst->next;
					DELETE( Ycurrent );
				}
				return 1;
			}
			Ycurrent->next = NULL;
			if ( Yfirst == NULL ) {
				Yfirst = Ycurrent;
				Ylast = Ycurrent;
			} else {
				Ylast->next = Ycurrent;
				Ylast = Ycurrent;
			}
		} else if ( Ycurrent != NULL )
			Ycurrent = Ycurrent->next;
		else {
			Ycurrent = seedY32;
			Yfirst = seedY32;
		}
		seed32to22 = (0x7ff0UL & (unsigned long)rand()) >> 17;
		seed21to11 = (0x7ff0UL & (unsigned long)rand()) << 6;
		seed10to1 = (0x7fe0UL & (unsigned long)rand()) >> 5;
		Ycurrent->random = (seed32to22 | seed21to11 | seed10to1);
		switch ( t ) {
			case 0 :
				Yt521 = Ycurrent;
				Yt17 = Ycurrent;
				break;
			case 1 :
				Yt16 = Ycurrent;
				break;
			case 15 :
				Yt15 = Ycurrent;
				break;
			case 16 :
				Yt1 = Ycurrent;
				break;
		}
	}

	/* modify t = 16 */
	Ycurrent->random = (Yfirst->random>>9 ^ Yt15->random) | (0xffffffffUL & Ycurrent->random<<23);

	/* 17 <= t <= 520 */
	for ( t = 19; t < 521; t++ ) {
		if ( seedY32 == NULL ) {
			Ycurrent = NEW( MaxSeqRandNum );
			if ( Ycurrent == NULL ) {
				while ( Yfirst != NULL ) {
					Ycurrent = Yfirst;
					Yfirst = Yfirst->next;
					DELETE( Ycurrent );
				}
				return 2;
			}
			Ycurrent->next = NULL;
			Ylast->next = Ycurrent;
			Ylast = Ycurrent;
		} else
			Ycurrent = Ycurrent->next;
		Ycurrent->random = 0xffffffffUL & ((Yt17->random<<23 | Yt16->random>>9) ^ Yt1->random);
		if ( t == 489 )
			Yt32 = Ycurrent;
		Yt17 = Yt16;
		Yt16 = Yt16->next;
		Yt1 = Yt1->next;
	}

	/* ring seedY32 if not ringed */
	if ( seedY32 == NULL ) {
		Ylast->next = Yfirst;
		seedY32 = Yfirst;
	}

	/* set Ycurrent */
	Ycurrent = Yfirst;

	/* make randX32 ring from seedY32 ring */
	for ( f = 0, s = 0, t = 0; t < 8321; f--, t++ ) {
		if ( f == 0 ) {
			if( randX32 == NULL ) {
				Xcurrent = NEW( MaxSeqRandNum );
				if ( Xcurrent == NULL ) {
					for ( t = 0; t < 521; t++ ) {
						Ycurrent = Yfirst;
						Yfirst = Yfirst->next;
						DELETE( Ycurrent );
					}
					while ( Xfirst != NULL ) {
						Xcurrent = Xfirst;
						Xfirst = Xfirst->next;
						DELETE( Xcurrent );
					}
					seedY32 = NULL;
					Xt521 = NULL;
					Xt32 = NULL;
					return 3;
				}
				Xcurrent->next = NULL;
				if ( Xfirst == NULL ) {
					Xfirst = Xcurrent;
					Xlast = Xcurrent;
				} else {
					Xlast->next = Xcurrent;
					Xlast = Xcurrent;
				}
			} else if ( Xcurrent != NULL )
				Xcurrent = Xcurrent->next;
			else {
				Xcurrent = randX32;
				Xfirst = randX32;
			}
			Xcurrent->random = Ycurrent->random;
			switch ( s ) {
				case 0 :
					Xt521 = Xcurrent;
					break;
				case 489 :
					Xt32 = Xcurrent;
					break;
			}
			s++;
			f = 16;
		}
		Ycurrent->random = Yt32->random ^ Yt521->random;
		Ycurrent = Ycurrent->next;
		Yt32 = Yt32->next;
		Yt521 = Yt521->next;
	}

	/* ring randX32 if not ringed */
	if ( randX32 == NULL ) {
		Xlast->next = Xfirst;
		randX32 = Xfirst;
	}

	return 0;
}

/*** Mrand32 ***/

unsigned long Mrand32( void )
{
	unsigned long rand;

	if ( randX32 == NULL )
		if ( Msrand32( 1u ) != 0 )	/* if no memory... */
			return 0UL;	/* return 0 */

	rand = randX32->random;

	randX32->random = (Xt32->random) ^ (Xt521->random);	/* generate next random */
	randX32 = randX32->next;
	Xt32 = Xt32->next;
	Xt521 = Xt521->next;

	return rand;
}

/* Mgaussrand32 */

double Mgaussrand32( double mu, double sigma )
{
	static long double V1, V2, S;
	static short phase = 0;
	long double X;

	if (phase == 0) {
		do {
			unsigned long r1, r2;
			long double U1, U2;
			r1 = Mrand32();
			r2 = Mrand32();
			if (r1 == 0 && r2 == 0)
				return 0;
			U1 = (long double)r1 / MRAND32_MAX;
			U2 = (long double)r2 / MRAND32_MAX;

			V1 = 2 * U1 - 1;
			V2 = 2 * U2 - 1;
			S = V1 * V1 + V2 * V2;
		} while (S >= 1 || S == 0);

		X = V1 * sqrt((double)(-2 * log((double)S) / S));
	} else
		X = V2 * sqrt((double)(-2 * log((double)S) / S));

	phase = 1 - phase;

	return (double)(X * sigma + mu);
}